// Scintilla source code edit control
/** @file LexAsm.cxx
 ** Lexer for MASM syntax Assembler
 ** Written by Mincho Georgiev (c) 2006
 ** ver. 1.0
 **/
// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>

#include "Platform.h"

#include "PropSet.h"
#include "Accessor.h"
#include "StyleContext.h"
#include "KeyWords.h"
#include "Scintilla.h"
#include "SciLexer.h"

#define MAIN_HEADER 5001
#define SMALL_HEADER 5002
#define ELSE_HEADER 5003
#define END_MAIN_HEADER 5004
#define END_SMALL_HEADER 5005
#define BREAK 5006


#define COMMENT_DIRECTIVE_FOUND 2222

/*--------------------------------------------------------------------
	The lexer has a Preprocessor WordList as well,
	that doesn't affect on the conditional assebly folding
	and on the repeat block folding also.
	The conditional assembly directives and
	the repeat block directives can be directives
	or preprocessor directives as well.
	The constant in SciLexer.h must be 
	defined as SCE_ASM_PREPROCESSOR 15
----------------------------------------------------------------------*/

static inline bool IsAWordChar(const int ch) {
	return (ch < 0x80) && (isalnum(ch) || ch == '.' ||
		ch == '_' || ch == '?');
}

static inline bool IsAWordStart(const int ch) {
	return (ch < 0x80) && (isalnum(ch) || ch == '_' || ch == '.' ||
		ch == '%' || ch == '@' || ch == '$' || ch == '?');
}

static inline bool IsAsmOperator(char ch) {
	if (isalnum(ch))
		return false;

	if (ch == '*' || ch == '/' || ch == '-' || ch == '+' ||
		ch == '(' || ch == ')' || ch == '=' || ch == '^' ||
		ch == '[' || ch == ']' || ch == '<' || ch == '&' ||
		ch == '>' || ch == ',' || ch == '|' || ch == '~' ||
		ch == '%' || ch == ':')
		return true;
	return false;
}

static inline void _getkeyword(Accessor &styler, unsigned int pos, char keywordbuffer[])
{
		
		for(unsigned int i = 0; i < 14; i++)
		{
			if(!iswordchar(styler[pos + i])){
				break;
			}
			keywordbuffer[i] = static_cast<char>(tolower(styler[pos + i]));
			keywordbuffer[i + 1] = '\0';
		}
}


static inline int _checkfoldpoint(char *keywstr)
{	
	
		if ((!strcmp(keywstr, "proc")) || !strcmp(keywstr, "struct") 
			|| !strcmp(keywstr, "macro") 
			|| !strcmp(keywstr,"union")
			|| !strcmp(keywstr, ".while")
			|| !strcmp(keywstr, ".repeat")){
			return MAIN_HEADER;
			
		}else if (!strcmp(keywstr, ".if")){
			return SMALL_HEADER;
			
		}else if (!strcmp(keywstr, ".elseif") || !strcmp(keywstr, ".else")){
			return ELSE_HEADER;
			
		}else if ((!strcmp(keywstr, "endp")) 
			|| !strcmp(keywstr, "endm") 
			|| !strcmp(keywstr, "ends")
			|| !strcmp(keywstr, ".until") 
			|| !strcmp(keywstr, ".endw")){
			return END_MAIN_HEADER;
			
		}else if (!strcmp(keywstr, ".endif")){
			return END_SMALL_HEADER;
			
		}else if(!strcmp(keywstr, ".break")){
			return BREAK;
		}
		
	return 0;
}
static inline int _conditionalassembly(char *keywstr)
{

		if(!strcmp(keywstr, "endif")){
			return END_SMALL_HEADER;
		}
		
		else if (!strcmp(keywstr, "ifdef")
		||!strcmp(keywstr, "if")
		|| !strcmp(keywstr, "ifidn")   
		|| !strcmp(keywstr, "ifndef")
		|| !strcmp(keywstr, "ife")){
			return SMALL_HEADER;
		}
		else if(!strcmp(keywstr, "else")
			|| !strcmp(keywstr, "elseif")
			|| !strcmp(keywstr, "elseifdef")
			|| !strcmp(keywstr, "elseifndef")){
				return ELSE_HEADER;
				
		}
	return 0;
}

static inline int _repeatblockfoldpoint(char *keywstr)
{

		if(!strcmp(keywstr, "repeat")
			|| !strcmp(keywstr, "while")
			|| !strcmp(keywstr, "forc")
			|| !strcmp(keywstr, "for")){
			return MAIN_HEADER;
		}
	return 0;
}

static void ColouriseAsmDoc(unsigned int startPos, int length, int initStyle, WordList *keywordlists[],
                            Accessor &styler) {

	WordList &cpuInstruction = *keywordlists[0];
	WordList &mathInstruction = *keywordlists[1];
	WordList &registers = *keywordlists[2];
	WordList &directive = *keywordlists[3];
	WordList &directiveOperand = *keywordlists[4];
	WordList &extInstruction = *keywordlists[5];
	WordList &preprocessor = *keywordlists[6];
	
	// Do not leak onto next line
	if (initStyle == SCE_ASM_STRINGEOL)
		initStyle = SCE_ASM_DEFAULT;
	StyleContext sc(startPos, length, initStyle, styler);
	for (; sc.More(); sc.Forward())
	{
		// Prevent SCE_ASM_STRINGEOL from leaking back to previous line
		if (sc.atLineStart && (sc.state == SCE_ASM_STRING)) {
			sc.SetState(SCE_ASM_STRING);
		} else if (sc.atLineStart && (sc.state == SCE_ASM_CHARACTER)) {
			sc.SetState(SCE_ASM_CHARACTER);
		}
		// Handle line continuation generically.
		if (sc.ch == '\\') {
			if (sc.chNext == '\n' || sc.chNext == '\r') {
				sc.Forward();
				if (sc.ch == '\r' && sc.chNext == '\n') {
					sc.Forward();
				}
				continue;
			}
		}
		// Determine if the current state should terminate.
		if (sc.state == SCE_ASM_OPERATOR) {
			if (!IsAsmOperator(static_cast<char>(sc.ch))) {
			    sc.SetState(SCE_ASM_DEFAULT);
			}
		}else if (sc.state == SCE_ASM_NUMBER) {
			if (!IsAWordChar(sc.ch)) {
				sc.SetState(SCE_ASM_DEFAULT);
			}
		} else if (sc.state == SCE_ASM_IDENTIFIER) {
			if (!IsAWordChar(sc.ch) ) {
				char s[100];
				sc.GetCurrentLowered(s, sizeof(s));
				if (cpuInstruction.InList(s)) {
					sc.ChangeState(SCE_ASM_CPUINSTRUCTION);
				} else if (mathInstruction.InList(s)) {
					sc.ChangeState(SCE_ASM_MATHINSTRUCTION);
				} else if (registers.InList(s)) {
					sc.ChangeState(SCE_ASM_REGISTER);
				}  else if (directive.InList(s)) {
					sc.ChangeState(SCE_ASM_DIRECTIVE);
				} else if (directiveOperand.InList(s)) {
					sc.ChangeState(SCE_ASM_DIRECTIVEOPERAND);
				} else if (extInstruction.InList(s)) {
					sc.ChangeState(SCE_ASM_EXTINSTRUCTION);
				} else if (preprocessor.InList(s)) {
					sc.ChangeState(SCE_ASM_PREPROCESSOR);
				}
				
				sc.SetState(SCE_ASM_DEFAULT);
				/* 'comment' directive detection */
				if (!strcmp(s, "comment")) {
					sc.SetState(COMMENT_DIRECTIVE_FOUND);
				}
				if (!strcmp(s, "comment") && sc.ch == '*'){
					sc.SetState(SCE_ASM_COMMENTBLOCK);
				}
				
			}
		}
		
		/*comment block delimiters*/
		else if (sc.state == COMMENT_DIRECTIVE_FOUND)
		{
			if(sc.ch == '*'){
				sc.SetState(SCE_ASM_COMMENTBLOCK);
				
			}
			else{
				while(!sc.atLineEnd && sc.ch != '*'){
					sc.Forward();
					
				}if(sc.ch == '*') {
						sc.SetState(SCE_ASM_COMMENTBLOCK);
						
				}else sc.SetState(SCE_ASM_DEFAULT);
			}
				
		}
		else if (sc.state == SCE_ASM_COMMENTBLOCK){
			if(sc.ch == '*'){
				sc.ForwardSetState(SCE_ASM_DEFAULT); 
			}
		}
		
		else if (sc.state == SCE_ASM_COMMENT ) {
			if (sc.atLineEnd) {
				sc.SetState(SCE_ASM_DEFAULT);
			}
		} else if (sc.state == SCE_ASM_STRING) {
			if (sc.ch == '\\') {
				if (sc.chNext == '\"' || sc.chNext == '\'' || sc.chNext == '\\') {
					sc.Forward();
				}
			} else if (sc.ch == '\"') {
				sc.ForwardSetState(SCE_ASM_DEFAULT);
			} else if (sc.atLineEnd) {
				sc.ChangeState(SCE_ASM_STRINGEOL);
				sc.ForwardSetState(SCE_ASM_DEFAULT);
			}
		} else if (sc.state == SCE_ASM_CHARACTER) {
			if (sc.ch == '\\') {
				if (sc.chNext == '\"' || sc.chNext == '\'' || sc.chNext == '\\') {
					sc.Forward();
				}
			} else if (sc.ch == '\'') {
				sc.ForwardSetState(SCE_ASM_DEFAULT);
			} else if (sc.atLineEnd) {
				sc.ChangeState(SCE_ASM_STRINGEOL);
				sc.ForwardSetState(SCE_ASM_DEFAULT);
			}
		}
		// Determine if a new state should be entered.
		if (sc.state == SCE_ASM_DEFAULT) {
			if (sc.ch == ';'){
				sc.SetState(SCE_ASM_COMMENT);
			} else if (isdigit(sc.ch) || (sc.ch == '.' && isdigit(sc.chNext))) {
				sc.SetState(SCE_ASM_NUMBER);
			} else if (IsAWordStart(sc.ch)) {
				sc.SetState(SCE_ASM_IDENTIFIER);
			} else if (sc.ch == '\"') {
				sc.SetState(SCE_ASM_STRING);
			} else if (sc.ch == '\'') {
				sc.SetState(SCE_ASM_CHARACTER);
			} else if (IsAsmOperator(static_cast<char>(sc.ch))) {
				sc.SetState(SCE_ASM_OPERATOR);
			}	
		}
	}
	sc.Complete();
}


static void FoldAsmDoc(unsigned int startPos, int length, int initStyle, WordList *[],
                            Accessor &styler) 
{	
	//fold property check:
	if(!styler.GetPropertyInt("fold")) return;
	bool foldComment = styler.GetPropertyInt("fold.comment") != 0;
	bool foldAtConditionalAssembly = styler.GetPropertyInt("fold.at.conditional.assembly") != 0;
	bool foldCompact = styler.GetPropertyInt("fold.compact", 1) != 0;
	bool foldAtElse = styler.GetPropertyInt("fold.at.else", 0) != 0;
	bool foldAtRepeatBlock = styler.GetPropertyInt("fold.at.repeat.block",0) != 0;
	
	unsigned int endPos = startPos + length;
	int visibleChars = 0;
	int lineCurrent = styler.GetLine(startPos);
	int levelCurrent = SC_FOLDLEVELBASE;
	if (lineCurrent > 0)
		levelCurrent = styler.LevelAt(lineCurrent-1) >> 16;
	int levelMinCurrent = levelCurrent;
	int levelNext = levelCurrent;
	int levelMinCurrentElse = levelCurrent;
	char chNext = styler[startPos];
	int styleNext = styler.StyleAt(startPos);
	int style = initStyle;
	bool break_directive_found = false;
	char keyword[14];
	
	
	
	for (unsigned int i = startPos; i < endPos; i++) 
	{
		char ch = static_cast<char>(tolower(chNext));
		chNext = static_cast<char>(tolower(styler.SafeGetCharAt(i + 1)));
		char chPrev = static_cast<char>(tolower(styler.SafeGetCharAt(i - 1)));
		int stylePrev = style;
		style = styleNext;
		styleNext = styler.StyleAt(i + 1);
		bool atEOL = (ch == '\r' && chNext != '\n') || (ch == '\n');
		
		
		
		if(style == SCE_ASM_DIRECTIVE)
		{
			_getkeyword(styler, i, keyword);
			switch(_checkfoldpoint(keyword))
			{
				case MAIN_HEADER:
					levelNext++;
					break;
					
				case SMALL_HEADER:
					if(!break_directive_found)
					{	if (levelMinCurrentElse > levelNext) {
							levelMinCurrentElse = levelNext;
						}
						levelNext++;
					}
					else if(break_directive_found)
					{
						break_directive_found = false;
					}
					break;
					
				case ELSE_HEADER:
					levelMinCurrentElse = levelNext - 1;
					break;
					
				case END_MAIN_HEADER:
				case END_SMALL_HEADER:	
					levelNext--;
					break;
					
				case BREAK: /*'.break'*/
					break_directive_found = true;
					break;
				
				
			}
		}
		//the repeat block directives can be a preprocessor keyword or either a directive ones:
		if(foldAtRepeatBlock && (style == SCE_ASM_PREPROCESSOR)
			|| foldAtRepeatBlock && (style == SCE_ASM_DIRECTIVE))
		{
			
			_getkeyword(styler, i, keyword);
			if(_repeatblockfoldpoint(keyword) == MAIN_HEADER && chPrev != '.'){
				levelNext++;
			}
		}
		
		
		/* comment block folding */		
		if(foldComment && style == SCE_ASM_COMMENTBLOCK)
		{
			if (levelMinCurrent > levelNext) {
				levelMinCurrent = levelNext;
			}
			//the level is incremented only on the 1st line
			//of the comment block:
			if(styler.StyleAt(i-1) != SCE_ASM_COMMENTBLOCK){	
				levelNext++;
			}
			//comment block ends:
			if(ch == '*' && styler.StyleAt(i + 1) == SCE_ASM_DEFAULT){
				levelNext--;
					
			}
			
		}
		
		if(foldAtConditionalAssembly && (style == SCE_ASM_PREPROCESSOR)
			|| foldAtConditionalAssembly && (style == SCE_ASM_DIRECTIVE))
		{	
		
			if((chPrev != '.') && (ch != '.') && (!iswordchar(chPrev)))
			{
				_getkeyword(styler, i, keyword);
				switch(_conditionalassembly(keyword))
				{
					case SMALL_HEADER:
						if (levelMinCurrent> levelNext) {
							levelMinCurrent = levelNext;
						}
						levelNext++;
						break;
					
					case ELSE_HEADER:
						if(foldAtElse){
							levelMinCurrentElse = levelNext - 1;
						}
						break;
					
					case END_SMALL_HEADER:
						levelNext--;
						break;
				}
			}
		}
		
		if (atEOL) 
		{
			int levelUse = levelCurrent;
			
			if (foldAtElse && (levelMinCurrentElse < levelUse)) {
				levelUse = levelMinCurrentElse;
			}

			int lev = levelUse | levelNext << 16;
			
			if (visibleChars == 0 && foldCompact)
				lev |= SC_FOLDLEVELWHITEFLAG;
			if (levelUse < levelNext)
				lev |= SC_FOLDLEVELHEADERFLAG;
			if (lev != styler.LevelAt(lineCurrent)) {
				styler.SetLevel(lineCurrent, lev);
			}
			lineCurrent++;
			levelCurrent = levelNext;
			levelMinCurrent = levelCurrent;
			 levelMinCurrentElse = levelCurrent;
			visibleChars = 0;
			
		}
		 if (!isspacechar(ch)) visibleChars++;
		 
	}
   
 
}

static const char * const asmWordListDesc[] = {
	"CPU instructions",
	"FPU instructions",
	"Registers",
	"Directives",
	"Directive operands",
	"Extended instructions",
	"Preprocessor directives",
	0
};


LexerModule lmAsm(SCLEX_ASM, ColouriseAsmDoc, "asm", FoldAsmDoc, asmWordListDesc);





